package hexagon.world;

import static org.lwjgl.glfw.GLFW.*;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.Random;
import java.util.function.Consumer;

import org.joml.Vector3f;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;

import gnu.trove.impl.Constants;
import gnu.trove.iterator.TLongIterator;
import gnu.trove.list.array.TLongArrayList;
import gnu.trove.map.hash.TLongIntHashMap;
import gnu.trove.map.hash.TLongObjectHashMap;
import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.model.RawModel;
import hexagon.util.MathUtil;
import hexagon.world.block.BlockType;
import hexagon.world.gen.ChunkGen;

public class World { // TODO: Move stuff from Chunk and Block to World, and make BlockTypes
	
	public static float viewDistance = 12f;
	public static float reachDistance = 8f;
	public static final double angleHalfHexagon = Math.PI / Chunk.totalSize;
	public static final double radius = HexagonModel.y60 / angleHalfHexagon;
	public static final float blockToCoordRatio = HexagonModel.y60 * 2;
	private static final int[] chunkOffsets = new int[6 * 3];
	static {
		for (int i = 0, idx = 0; i < 6; i++) {
			chunkOffsets[idx++] = i % 3 == 0 ? i / 3 * 2 - 1 : 0;
			chunkOffsets[idx++] = (i + 2) % 3 == 0 ? (i + 2) / 3 * 2 - 3 : 0;
			chunkOffsets[idx++] = (i + 1) % 3 == 0 ? (i + 1) / 3 * 2 - 3 : 0;
		}
	}
	public final long seed;
	protected ChunkGen landscapeGen;
	
	private int ticks = 0;
	
	private TLongObjectHashMap<Cluster> clusters = new TLongObjectHashMap<>();
	private TLongArrayList blocksToUpdate = new TLongArrayList();
	private TLongArrayList blocksUpdating = new TLongArrayList();
	private TLongArrayList chunksToUpdate = new TLongArrayList();
	private TLongArrayList chunksUpdating = new TLongArrayList();
	private long hoverBlock;
	private boolean hoverBlockExists;
	private ByteBuffer hoverBlockRenderBuffer;
	private int rightClickCountdown = 0, leftClickCountdown = 0;
	private boolean leftDownPrev, rightDownPrev;
	
	// Rendering blocks
	private int[] blockRenderStrides = new int[8];
	private ByteBuffer[] blockRenderBuffers = new ByteBuffer[8];
	private TLongIntHashMap[] blockToIndexMaps = new TLongIntHashMap[8];
	private long[][] indexToBlockLists = new long[8][];
	
	private RawModel skyModel = Loader.theLoader.loadToVAO(new float[] { -1, -1, 1, -1, -1, 1, 1, 1 });
	private Vector3f sunDir = new Vector3f();
	
	// Saving to file
	protected File saveFolder;
	protected ByteBuffer saveBuffer = ByteBuffer.allocate(1024);
	protected ByteBuffer readBuffer = ByteBuffer.allocate(1024);
//	protected long bytes = 0, time = 0;
	protected Consumer<FileChannel> saveBufferFull = (fc) -> {
		try {
//			bytes += saveBuffer.position();
			saveBuffer.flip();
//			long start = System.nanoTime();
			fc.write(saveBuffer);
//			time += System.nanoTime() - start;
			saveBuffer.clear();
		} catch (IOException e) {
			e.printStackTrace();
		}
	};
	
	public World() {
		seed = new Random().nextLong();
		landscapeGen = new ChunkGen(this, new Random(seed));
		
		for (int i = 0; i < 8; i++) {
			BlockRenderer<Block> renderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
			blockRenderStrides[i] = renderer.instanceSize;
			blockRenderBuffers[i] = ByteBuffer.allocate(renderer.numberOfInstances * blockRenderStrides[i]).order(ByteOrder.nativeOrder());
			blockToIndexMaps[i] = new TLongIntHashMap(Constants.DEFAULT_CAPACITY, Constants.DEFAULT_LOAD_FACTOR, 0, -1);
			indexToBlockLists[i] = new long[renderer.numberOfInstances];
		}
		hoverBlockRenderBuffer = ByteBuffer.allocate(Math.max(BlockRenderer.block.instanceSize, BlockRenderer.side.instanceSize)).order(ByteOrder.nativeOrder());
	}
	
	public void tick(Camera camera) {
		updateHoverBlock();
		updateBlocks();
		
		double angle = camera.position.z * 2 * Math.PI / Chunk.totalSize;
		sunDir.set(0.5f, (float) Math.cos(ticks * 0.00001f - angle), (float) Math.sin(ticks * 0.00001f - angle));
		
		ticks++;
	}
	
	// TODO: Rewrite to be smarter!
	public void updateLoadedChunks(Camera camera) {
		Vector3f cVec = new Vector3f(), rVec = new Vector3f(); // chunk, real
		float camX = rVec.x = camera.position.x / 16;
		float camY = rVec.y = camera.position.y / 16;
		float camZ = rVec.z = MathUtil.superMod(camera.position.z / 16, Chunk.ringSize * 2 * HexagonModel.y60);
		MathUtil.realCoordsToChunk(rVec, cVec);
		
		if (getChunk((int) Math.floor(cVec.x), (int) Math.floor(cVec.y) - 1, (int) Math.floor(cVec.z)) == null) {
			setNewChunk((int) Math.floor(cVec.x), (int) Math.floor(cVec.y) - 1, (int) Math.floor(cVec.z));
		} else {
			TLongArrayList tempList = chunksUpdating;
			chunksUpdating = chunksToUpdate;
			chunksToUpdate = tempList;
			
			for (TLongIterator it = chunksUpdating.iterator(); it.hasNext();) {
				long key = it.next();
				Cluster cluster = clusters.get(key >>> 12);
				if (cluster == null) continue;
				Chunk c = cluster.getChunk((int) (key & 0xfff));
				if (c == null) continue;
				
				int cx = c.x(), cy = c.y(), cz = c.z(), ix, iy, iz;
				cVec.x = cx + 0.5f;
				cVec.y = cy + 0.5f;
				cVec.z = cz + 0.5f;
				MathUtil.chunkCoordsToReal(cVec, rVec);
				
				float xx, yy, zz, distSq;
				xx = rVec.x - camX;
				yy = rVec.y - camY;
				zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
				xx /= 1.5f;
				zz /= blockToCoordRatio;
				distSq = xx * xx + yy * yy + zz * zz;
				distSq /= blockToCoordRatio * blockToCoordRatio;
				
				if (distSq > (viewDistance + 1) * (viewDistance + 1)) {
					cluster.removeChunk(cx, cy, cz);
					Chunk ch = null;
					int idx = 0;
					for (int i = 0; i < 6; i++) {
						ix = cx + chunkOffsets[idx++];
						iy = cy + chunkOffsets[idx++];
						iz = cz + chunkOffsets[idx++];
						iz = MathUtil.superMod(iz, Chunk.ringSize);
						
						long iCoords = encodeClusterCoords(ix >> 4, iy >> 4, iz >> 4);
						Cluster theCluster = iCoords == cluster.coords ? cluster : clusters.get(iCoords);
						if (theCluster != null && (ch = theCluster.getChunk(ix, iy, iz)) != null) {
							chunksToUpdate.add(ch.worldCoords());
						}
					}
				} else {
					boolean stillNeedsUpdate = false;
					int idx = 0;
					for (int i = 0; i < 6; i++) {
						ix = cx + chunkOffsets[idx++];
						iy = cy + chunkOffsets[idx++];
						iz = cz + chunkOffsets[idx++];
						iz = MathUtil.superMod(iz, Chunk.ringSize);
						
						long iCoords = encodeClusterCoords(ix >> 4, iy >> 4, iz >> 4);
						Cluster theCluster = iCoords == cluster.coords ? cluster : clusters.get(iCoords);
						if (theCluster == null || theCluster.getChunk(ix, iy, iz) == null) {
							cVec.x = ix + 0.5f;
							cVec.y = iy + 0.5f;
							cVec.z = iz + 0.5f;
							MathUtil.chunkCoordsToReal(cVec, rVec);
							xx = rVec.x - camX;
							yy = rVec.y - camY;
							zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
							xx /= 1.5f;
							zz /= blockToCoordRatio;
							distSq = xx * xx + yy * yy + zz * zz;
							if (distSq <= viewDistance * viewDistance) {
								setNewChunk(ix, iy, iz);
							} else stillNeedsUpdate = true;
						}
					}
					if (stillNeedsUpdate) chunksToUpdate.add(key);
				}
				if (cluster.size() <= 0) clusters.remove(cluster.coords);
			}
			
			chunksUpdating.reset();
		}
	}
	
	private void updateHoverBlock() {
		Main.mousePicker.updateBlockPointedAt(this);
		hoverBlock = Main.mousePicker.getHoverBlock();
		hoverBlockExists = Main.mousePicker.hoverBlockExists();
		
		if (hoverBlockExists) {
			boolean left = glfwGetMouseButton(Main.main.window, GLFW_MOUSE_BUTTON_LEFT) == GLFW_PRESS, right = glfwGetMouseButton(Main.main.window, GLFW_MOUSE_BUTTON_RIGHT) == GLFW_PRESS;
			
			Chunk chunk = getChunk(hoverBlock >>> 12);
			short bCoords = (short) (hoverBlock & 0xfff);
			int hx = chunk.bx(bCoords), hy = chunk.by(bCoords), hz = chunk.bz(bCoords);
			if (left) {
				if (!right) {
					if (leftClickCountdown == 0 || !leftDownPrev) {
						removeBlock(hx, hy, hz);
						hoverBlock = 0;
						hoverBlockExists = false;
						leftClickCountdown = 12;
					}
				}
			} else if (right) {
				if (rightClickCountdown == 0 || !rightDownPrev) {
					int side = Main.mousePicker.getHoverSide();
					if (side != -1) {
						int xx = hx + Block.neighbourOffsets[side][0];
						int yy = hy + Block.neighbourOffsets[side][1];
						int zz = hz + Block.neighbourOffsets[side][2];
						zz = MathUtil.superMod(zz, Chunk.totalSize);
						BlockType block = getBlock(xx, yy, zz);
						Vector3f bc = Main.main.camera.blockCoords;
						if (block == null && !(bc.x == xx && bc.y == yy && bc.z == zz)) {
							setBlock(xx, yy, zz, Main.main.currentType);
						}
					}
					rightClickCountdown = 12;
				}
			}
			leftDownPrev = left;
			rightDownPrev = right;
		}
		
		if (leftClickCountdown > 0) leftClickCountdown--;
		if (rightClickCountdown > 0) rightClickCountdown--;
	}
	
	public void render(Camera camera) {
		Shader.sky.enable();
		Shader.sky.setUniformMat4("invProjMatr", camera.invProjMatr);
		Shader.sky.setUniformMat4("invViewMatr", camera.invViewMatr);
		Shader.sky.setUniform3f("sun", sunDir.x, sunDir.y, sunDir.z);
		GL30.glBindVertexArray(skyModel.getVaoID());
		GL11.glDisable(GL11.GL_DEPTH_TEST);
		GL11.glDrawArrays(GL11.GL_TRIANGLE_STRIP, 0, skyModel.getVertexCount());
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		
		Shader.block.enable();
		Shader.block.setUniform3f("sunDirection", sunDir.x, sunDir.y, sunDir.z);
		
		BlockRenderer<Block> blockRenderer;
//		long t1 = 0, tmp1;
		for (int i = 0; i < 8; i++) {
			if ((Main.sides & 1 << i) == 0) continue;
			Shader.block.setUniform1i("side", i);
			blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
			blockRenderer.begin();
			blockRenderer.submitPerVertexData(i);
//			tmp1 = System.nanoTime();
			
			blockRenderer.submit(blockRenderBuffers[i]);
			
//			Block b;
//			for (Chunk c : chunks.values()) {
//				for (Long l : c.blockRendered.get(i)) {
//					b = c.blocks.get(l);
//					if ((b.sidesToRender & 1 << i) == 0) System.out.println("oops!!");
//					if (b != hoverBlock) blockRenderer.submit(b);
//					num++;
//				}
//			}
//			t1 += System.nanoTime() - tmp1;
//			blockRenderer.end();
//			blockRenderer.render();
		}
//		System.out.println(t1);
		
		if (hoverBlockExists) {
			Chunk chunk = getChunk(hoverBlock >>> 12);
			int bCoords = (int) (hoverBlock & 0xfff);
			hoverBlockRenderBuffer.clear();
			submitBlock(hoverBlock, hoverBlockRenderBuffer);
			Shader.block.setUniform1f("glow", 1.5f);
			for (int i = 0; i < 8; i++) {
				Shader.block.setUniform1i("side", i);
				if ((Main.sides & 1 << i) == 0 || (chunk.sidesToRender[bCoords] & 1 << i) == 0) continue;
				blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
				blockRenderer.begin();
				blockRenderer.submitPerVertexData(i);
				blockRenderer.submit(hoverBlockRenderBuffer);
				blockRenderer.end();
				blockRenderer.render();
			}
			Shader.block.setUniform1f("glow", 1f);
		}
		
		if (Main.main.moveWithMouse) {
			GL20.glUseProgram(0);
			
			GL30.glBindVertexArray(Main.main.camera.crosshair.getVaoID());
			GL11.glDrawArrays(GL11.GL_LINES, 0, 4);
		}
	}
	
	public void submitBlock(long b, ByteBuffer buffer) {
		Chunk c = getChunk(b >>> 12);
		short bCoords = (short) (b & 0xfff);
		BlockType type = c == null ? null : c.getBlock(bCoords);
		if (type == null) {
			buffer.position(buffer.position() + 16);
			return;
		}
		buffer.putInt(c.bx(bCoords)).putInt(c.by(bCoords)).putInt(c.bz(bCoords));
		buffer.putInt(type.texCoords);
	}
	
	protected void addToRenderList(int list, long block) {
		int N = blockToIndexMaps[list].size();
		if (blockToIndexMaps[list].containsKey(block)) {
			System.out.println("Oh!\t" + Long.toHexString(block));
			return;
		}
		
		blockToIndexMaps[list].put(block, N);
		if (N == indexToBlockLists[list].length) {
			long[] newArr = new long[N * 3 / 2];
			System.arraycopy(indexToBlockLists[list], 0, newArr, 0, N);
			indexToBlockLists[list] = newArr;
			
			ByteBuffer newBuffer = ByteBuffer.allocate(indexToBlockLists[list].length * blockRenderStrides[list]).order(ByteOrder.nativeOrder());
			System.arraycopy(blockRenderBuffers[list].array(), 0, newBuffer.array(), 0, blockRenderBuffers[list].capacity());
			newBuffer.position(blockRenderBuffers[list].position());
			blockRenderBuffers[list] = newBuffer;
		}
		indexToBlockLists[list][N] = block;
		
		submitBlock(block, blockRenderBuffers[list]);
	}
	
	protected void removeFromRenderList(int list, long block) {
		int index = blockToIndexMaps[list].remove(block);
		if (index == -1) return;
		int Nm1 = blockToIndexMaps[list].size(); // N - 1
		long last = indexToBlockLists[list][Nm1];
		int stride = blockRenderStrides[list];
		if (index != Nm1) {
			indexToBlockLists[list][index] = last;
			blockToIndexMaps[list].put(last, index);
			byte[] arr = blockRenderBuffers[list].array();
			try {
				System.arraycopy(arr, Nm1 * stride, arr, index * stride, stride);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		blockRenderBuffers[list].position(blockRenderBuffers[list].position() - stride);
	}
	
	private void updateBlocks() {
		TLongArrayList tempList = blocksUpdating;
		blocksUpdating = blocksToUpdate;
		blocksToUpdate = tempList;
		
		long b;
		Chunk chunk;
		BlockType block;
		short bCoords;
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			updateBlock(b);
		}
		
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			chunk = getChunk(b >>> 12);// >> 
			if (chunk != null) {
				block = chunk.getBlock(bCoords = (short) (b & 0xfff));
				if (block != null) chunk.willUpdate.clear(bCoords);
			}
		}
		
		blocksUpdating.reset();
	}
	
	public void updateBlock(int x, int y, int z) {
		updateBlock(World.encodeBlockCoords(x, y, z));
	}
	
	public void updateBlock(long coords) {
		Chunk chunk = getChunk(coords >>> 12);
		if (chunk == null) {
//			System.out.println("chunk == null" + Long.toHexString(coords));
			return;
		}
		short bCoords = (short) (coords & 0xfff);
		byte diff = chunk.sidesToRender[bCoords], after;
		chunk.onUpdateBlock(bCoords);
		diff ^= after = chunk.sidesToRender[bCoords];
		if (diff != 0) {
			for (int i = 0; i < 8; i++) {
				if ((diff & 1 << i) != 0) {
					if ((after & 1 << i) != 0) addToRenderList(i, coords);
					else removeFromRenderList(i, coords);
				}
			}
		}
	}
	
	public int getTerrainHeight(int x, int z) {
		return landscapeGen.getHeight(x, z) + 1;
	}
	
	public BlockType getBlock(long block) {
		Chunk c = getChunk(block >>> 12);
		return c == null ? null : c.getBlock((short) (block & 0xfff));
	}
	
	public BlockType getBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		Chunk chunk = getChunk(x >> 4, y >> 4, z >> 4);
		return chunk == null ? null : chunk.getBlock(x, y, z);
	}
	
	public void setBlock(long coords, BlockType block) {
		getChunkForSure(coords >>> 12).setBlock((short) (coords & 0xfff), block);
	}
	
	public void setBlock(int x, int y, int z, BlockType block) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		setBlock(encodeBlockCoords(x, y, z), block);
	}
	
	public void removeBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, 16), cy = Math.floorDiv(y, 16), cz = z / 16;
		
		Chunk chunk = getChunk(cx, cy, cz);
		if (chunk != null) chunk.removeBlock(x, y, z);
	}
	
	public Chunk getChunk(long chunk) {
		Cluster cluster = clusters.get(chunk >>> 12);
		return cluster != null ? cluster.getChunk((int) (chunk & 0xfff)) : null;
	}
	
	public Chunk getChunk(int x, int y, int z) {
		Cluster cluster = clusters.get(encodeClusterCoords(x >> 4, y >> 4, z >> 4));
		return cluster != null ? cluster.getChunk(Chunk.encodeCoords(x, y, z)) : null;
	}
	
	public Chunk getChunkForSure(long chunkCoordsIn) {
		long coords = chunkCoordsIn >>> 12;
		short chunkCoords = (short) (chunkCoordsIn & 0xfff);
		Cluster cluster = clusters.get(coords);
		Chunk chunk = null;
		if (cluster == null) {
			cluster = new Cluster(coords, this);
			clusters.put(coords, cluster);
		} else chunk = cluster.getChunk(chunkCoords);
		
		if (chunk == null) {
			chunk = new Chunk(chunkCoords, cluster);
			cluster.setChunk(chunk);
			
			chunksToUpdate.add(chunk.worldCoords());
		}
		
		return chunk;
	}
	
	public Chunk getChunkForSure(int x, int y, int z) {
		z = MathUtil.fitChunkZ(z);
		int cx = x >> 4, cy = y >> 4, cz = z >> 4;
		long coords = encodeClusterCoords(cx, cy, cz) << 12 | Cluster.encodeCoords(x, y, z);
		return getChunkForSure(coords);
	}
	
	public Chunk setNewChunk(int x, int y, int z) {
		return setNewChunk(x, y, z, false);
	}
	
	public Chunk setNewChunk(int x, int y, int z, boolean genNew) {
		z = MathUtil.fitChunkZ(z);
		int cx = x >> 4, cy = y >> 4, cz = z >> 4;
		long coords = encodeClusterCoords(cx, cy, cz);
		Cluster cluster = clusters.get(coords);
		if (cluster == null) {
			cluster = new Cluster(cx, cy, cz, this);
			clusters.put(coords, cluster);
		}
		Chunk chunk = new Chunk(x, y, z, cluster, genNew);
		cluster.setChunk(chunk);
		
		chunksToUpdate.add(chunk.worldCoords());
		return chunk;
	}
	
	public static long encodeClusterCoords(int x, int y, int z) { // bytes in long: xxxxyyzzzz
		return (x & 0xffffL) << 24 | (y & 0xffL) << 16 | MathUtil.superMod(z, Chunk.clusterCircumference) & 0xffffL;
	}
	
	public static long encodeBlockCoords(int x, int y, int z, short chunk, short block) { // bytes in long: xxxxyyzzzzcccbbb where b is blockcoords xyz and c is chunk
		return encodeClusterCoords(x, y, z) << 24 | chunk << 12 | block;
	}
	
	public static long encodeBlockCoords(long cluster, short chunk, short block) { // bytes in long: xxxxyyzzzzcccbbb where b is blockcoords xyz and c is chunk
		return cluster << 24 | chunk << 12 | block;
	}
	
	public static long encodeBlockCoords(int x, int y, int z) { // bytes in long: xxxxyyzzzzcccbbb where b is blockcoords xyz and c is chunk
		return encodeBlockCoords(x >> 8, y >> 8, z >> 8, Cluster.encodeCoords(x >> 4, y >> 4, z >> 4), Chunk.encodeCoords(x, y, z));
	}
	
	public void addBlockToUpdateList(Chunk c, short bCoords) {
		if (c == null) return;
		if (c.willUpdate == null) if (c.willUpdate.get(bCoords)) return;
		blocksToUpdate.add(c.worldCoords() << 12 | bCoords);
		c.willUpdate.set(bCoords);
	}
	
	public void addBlockToUpdateList(long coords) {
		addBlockToUpdateList(getChunk(coords >>> 12), (short) (coords & 0xfff));
	}
	
	public void setSaveFolder(File folder) {
		if (!folder.exists()) {
			folder.mkdir();
		}
		saveFolder = folder;
	}
	
	public void saveWorld() {
		if (saveFolder != null) {
			clusters.forEachValue(c -> {
				c.saveChunks();
				return true;
			});
		}
	}
	
	public void unloadWorld() {
		updateBlocks();
		
		clusters.forEachValue(c -> {
			c.onRemove(true);
			return true;
		});
		
		blocksToUpdate.clear();
		
		blocksToUpdate = null;
		clusters = null;
	}
	
}
