package hexagon;

import static org.lwjgl.glfw.Callbacks.*;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.system.MemoryStack.*;

import java.io.File;
import java.nio.DoubleBuffer;
import java.nio.IntBuffer;

import org.joml.Vector2d;
import org.joml.Vector2i;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;
import org.lwjgl.system.Configuration;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.system.MemoryUtil;

import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.util.MousePicker;
import hexagon.world.Chunk;
import hexagon.world.World;
import hexagon.world.block.BlockType;

public class Main {
	
	public static MousePicker mousePicker;
	public static Main main;
	
	private World world;
	public Camera camera;
	
	public boolean moveWithMouse = false;
	public static int sides = 0xff; // Sides to render, used for debugging
	public BlockType currentType;
	
	// The window handle
	public long window;
	private boolean isFullscreen = false;
	public Vector2i windowSize = new Vector2i(920, 540);
	private Vector2i windowSizeBeforeFullscreen = new Vector2i();
	private Vector2i windowPosBeforeFullscreen = new Vector2i();
	public Vector2d mousePos = new Vector2d();
	public Vector2d prevMousePos = new Vector2d();
	private DoubleBuffer xposArr = MemoryUtil.memAllocDouble(1), yposArr = MemoryUtil.memAllocDouble(1);
	
	private Main() {
		main = this;
	}
	
	public void run() {
		initWindow();
		initGame();
		loop();
		
		destroy();
		
		// Free the window callbacks and destroy the window
		glfwFreeCallbacks(window);
		glfwDestroyWindow(window);
		
		// Terminate GLFW and free the error callback
		glfwTerminate();
		glfwSetErrorCallback(null).free();
	}
	
	private void loop() {
		long prevTime = System.nanoTime(), currentTime, realPrevTime = prevTime, realDeltaTime = 0;
		int delta, ticks = 0, frames = 0, fps = 0;
		
		while (!glfwWindowShouldClose(window)) {
			delta = (int) (((currentTime = System.nanoTime()) - prevTime) * 1e-9 * 60);
			realPrevTime = currentTime;
			for (int i = 0; i < delta; i++) {
				if (ticks % 10 == 0) world.updateLoadedChunks(camera);
				tick();
				ticks++;
				if (ticks % 60 == 0) {
					fps = frames;
					frames = 0;
				}
				prevTime += 1e9 / 60;
			}
			//long tempTime = System.nanoTime() - realPrevTime;
			GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
			
			render();
			
			int error = GL11.glGetError();
			if (error != GL11.GL_NO_ERROR) System.out.println("OpenGL error: " + error);
			
			frames++;
			realDeltaTime = System.nanoTime() - realPrevTime;//tempTime;
			
			glfwSwapBuffers(window); // swap the color buffers
			
			String debugInfo = "    P: (" + camera.position.x + ", " + camera.position.y + ", " + camera.position.z + ")    C: (" + ((int) camera.blockCoords.x >> 4) + ", " + ((int) camera.blockCoords.y >> 4) + ", " + ((int) camera.blockCoords.z >> 4) + ")    R: (" + (float) (camera.rotation.x * 180 / Math.PI) + ", " + (float) (camera.rotation.y * 180 / Math.PI) + ", " + (float) (camera.rotation.z * 180 / Math.PI) + ")    v=" + (int) (100 * World.viewDistance) / 100f + (sides != 0xff ? "    s=" + Integer.toBinaryString(sides) : "");
			glfwSetWindowTitle(window, fps + " fps   ms: " + ((int) (realDeltaTime * 1e-6) < 10 ? "0" : "") + (int) (realDeltaTime * 1e-6) + debugInfo);
			
			// Poll for window events. The key callback above will only be
			// invoked during this call.
			glfwPollEvents();
		}
		
		world.saveWorld();
	}
	
	private void render() {
		world.render(camera);
		if (Main.main.moveWithMouse) {
			Shader.crosshair.enable();
			
			GL30.glBindVertexArray(Main.main.camera.crosshair.getVaoID());
			GL11.glDisable(GL11.GL_DEPTH_TEST);
			GL11.glDrawArrays(GL11.GL_LINES, 0, 4);
			GL11.glEnable(GL11.GL_DEPTH_TEST);
		}
	}
	
	private void tick() {
		glfwGetCursorPos(window, xposArr, yposArr);
		mousePos.set(xposArr.get(0), windowSize.y - yposArr.get(0));
		
		if (glfwGetKey(window, GLFW_KEY_MINUS) == GLFW_PRESS) World.viewDistance += 0.01f;// + unfortunately US-keyboard
		if (glfwGetKey(window, GLFW_KEY_SLASH) == GLFW_PRESS) World.viewDistance -= 0.01f;// -
		
		camera.update();
		mousePicker.update();
		world.tick(camera);
	}
	
	private void processKeys(long window, int key, int scancode, int action, int mods) {//action: 0 = release, 1 = press, 2 = repeat
		if (key == GLFW_KEY_ESCAPE && action == GLFW_RELEASE) glfwSetWindowShouldClose(window, true); // We will detect this in the rendering loop
		
		if (action == GLFW_PRESS) {
			boolean deletePressed = glfwGetKey(window, GLFW_KEY_DELETE) == GLFW_PRESS;
			
			switch (key) {
			// Delete-key required
			case GLFW_KEY_P:
				if (deletePressed) {
					camera.setPosition(0, world.getTerrainHeight(0, 0) + 3.2f, 0);
				}
				break;
			case GLFW_KEY_R:
				if (deletePressed) {
					camera.rotation.x = camera.rotation.y = camera.rotation.z = 0;
				}
				break;
			case GLFW_KEY_L:
				if (deletePressed) {
					world.unloadWorld();
					world = new World();
					camera = new Camera(window, 70f, 16f / 9f, 0.02f, 1000f, world);
					camera.position.y = world.getTerrainHeight(0, 0) + 3.5f;
					camera.loadProjMatrToShader(Shader.block);
					mousePicker = new MousePicker(camera);
					world.setSaveFolder(new File(System.getenv("appdata") + "/Hexagon/saves/TempWorld/"));
				}
				break;
			case GLFW_KEY_C:
				if (deletePressed) {
					world.setNewChunk((int) camera.blockCoords.x >> 4, (int) camera.blockCoords.y >> 4, (int) camera.blockCoords.z >> 4, true);
				}
				break;
				
				// The rest
			case GLFW_KEY_T:
				if (currentType == BlockType.grass) currentType = BlockType.sand;
				else currentType = BlockType.grass;
				break;
			case GLFW_KEY_M:
				moveWithMouse = !moveWithMouse;
				if (moveWithMouse) glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
				else glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
				break;
			case GLFW_KEY_F3:
				Shader.init();
				camera.loadProjMatrToShader(Shader.block);
				Shader.block.setUniform1i("blocksPerRow", 256 / 32);
				Shader.block.setUniform1f("radius", (float) World.radius);
				Shader.block.setUniform1f("totalSize", Chunk.totalSize);
				Shader.sky.enable();
				Shader.sky.setUniform2f("windowSize", windowSize.x, windowSize.y);
				Loader.theLoader.reloadTextures();
				System.out.println("Reloaded textures and shaders.");
				break;
			case GLFW_KEY_F11:
				if (!isFullscreen) {
					windowSizeBeforeFullscreen.set(windowSize);
					int[] winX = new int[1], winY = new int[1];
					glfwGetWindowPos(window, winX, winY);
					windowPosBeforeFullscreen.set(winX[0], winY[0]);
					prevMousePos.add(windowPosBeforeFullscreen.x, windowPosBeforeFullscreen.y);
					mousePos.add(windowPosBeforeFullscreen.x, windowPosBeforeFullscreen.y);
					long monitor = glfwGetPrimaryMonitor();
					GLFWVidMode mode = glfwGetVideoMode(monitor);
					System.out.println(mode.width() + "\t" + mode.height());
					glfwSetWindowMonitor(window, monitor, 0, 0, mode.width(), mode.height(), mode.refreshRate());
					isFullscreen = true;
				} else {
					prevMousePos.sub(windowPosBeforeFullscreen.x, windowPosBeforeFullscreen.y);
					mousePos.sub(windowPosBeforeFullscreen.x, windowPosBeforeFullscreen.y);
					glfwSetWindowMonitor(window, 0, windowPosBeforeFullscreen.x, windowPosBeforeFullscreen.y, windowSizeBeforeFullscreen.x, windowSizeBeforeFullscreen.y, GLFW_DONT_CARE);
					isFullscreen = false;
				}
				break;
			}
			if (key >= GLFW_KEY_0 && key <= GLFW_KEY_7) sides |= 1 << key - GLFW_KEY_0;
			else if (key == GLFW_KEY_8) sides = 0xff;
			else if (key == GLFW_KEY_9) sides = 0;
		}
	}
	
	private void initGame() {
		HexagonModel.init();
		Shader.init();
		currentType = BlockType.grass;
		
		world = new World();
		world.setSaveFolder(getAppDataFile(".hexagon/saves/TempWorld/"));
		
		camera = new Camera(window, 70f, 16f / 9f, 0.02f, 1000f, world);
		camera.position.y = world.getTerrainHeight(0, 0) + 3.2f;
		camera.loadProjMatrToShader(Shader.block);
		Shader.block.setUniform1i("blocksPerRow", 256 / 32);
		Shader.block.setUniform1f("radius", (float) World.radius);
		Shader.block.setUniform1f("totalSize", Chunk.totalSize);
		Shader.sky.enable();
		Shader.sky.setUniform2f("windowSize", windowSize.x, windowSize.y);
		Shader.crosshair.enable();
		Shader.crosshair.setUniform2f("windowSize", windowSize.x, windowSize.y);
		mousePicker = new MousePicker(camera);
	}
	
	private void initWindow() {
		// Setup an error callback. The default implementation
		// will print the error message in System.err.
		GLFWErrorCallback.createPrint(System.err).set();
		
		// Initialize GLFW. Most GLFW functions will not work before doing this.
		if (!glfwInit()) throw new IllegalStateException("Unable to initialize GLFW");
		
		// Configure GLFW
		glfwDefaultWindowHints(); // optional, the current window hints are already the default
		glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE); // the window will stay hidden after creation
		glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // the window will be resizable
		glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
		glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
		glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL11.GL_TRUE);
		glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
		glfwWindowHint(GLFW_SAMPLES, 1);
		
		// Create the window
		window = glfwCreateWindow(windowSize.x, windowSize.y, "Hexagon", 0, 0);
		if (window == 0) throw new RuntimeException("Failed to create the GLFW window");
		
		// Setup a key callback. It will be called every time a key is pressed, repeated or released.
		glfwSetKeyCallback(window, this::processKeys);
		
		glfwSetWindowSizeCallback(window, (window, width, height) -> {
			if (width != windowSize.x || height != windowSize.y) {
				GL11.glViewport(0, 0, width, height);
				camera.aspect = (float) width / height;
				camera.updateProjMatrix();
				camera.loadProjMatrToShader(Shader.block);
				Shader.sky.enable();
				Shader.sky.setUniform2f("windowSize", width, height);
				Shader.crosshair.enable();
				Shader.crosshair.setUniform2f("windowSize", width, height);
			}
			windowSize.set(width, height);
		});
		
		// Get the thread stack and push a new frame
		try (MemoryStack stack = stackPush()) {
			IntBuffer pWidth = stack.mallocInt(1); // int*
			IntBuffer pHeight = stack.mallocInt(1); // int*
			
			// Get the window size passed to glfwCreateWindow
			glfwGetWindowSize(window, pWidth, pHeight);
			
			// Get the resolution of the primary monitor
			GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());
			
			// Center the window
			glfwSetWindowPos(window, (vidmode.width() - pWidth.get(0)) / 2, (vidmode.height() - pHeight.get(0)) / 2);
		} // the stack frame is popped automatically
		
		// Make the OpenGL context current
		glfwMakeContextCurrent(window);
		// Enable v-sync
//		glfwSwapInterval(1);
		
		// Make the window visible
		glfwShowWindow(window);
		
		initGL();
	}
	
	private void initGL() {
		// This line is critical for LWJGL's interoperation with GLFW's
		// OpenGL context, or any context that is managed externally.
		// LWJGL detects the context that is current in the current thread,
		// creates the GLCapabilities instance and makes the OpenGL
		// bindings available for use.
		GL.createCapabilities();
		
//		Display.create(new PixelFormat().withSamples(2).withDepthBits(24), new ContextAttribs(3, 3));//samples = 2 or 4
		
//		GL11.glClearColor(0.1f, 0.2f, 0.7f, 1);
//		GL11.glEnable(GL13.GL_MULTISAMPLE);
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		GL11.glDepthFunc(GL11.GL_LEQUAL);
		
//		System.out.println(GL11.glGetInteger(GL12.GL_MAX_ELEMENTS_VERTICES));
	}
	
	public static void destroy() {
		Loader.theLoader.cleanUp();
		Shader.cleanUp();
/*		Mouse.destroy();
		Keyboard.destroy();
		Display.destroy();*/
	}
	
	public File getAppDataFile(String path) {
		String os = System.getProperty("os.name").toLowerCase();
		String appDataDir;
		if (os.contains("win")) {
			appDataDir = System.getenv("appdata");
		} else if (os.contains("mac")) {
			appDataDir = System.getProperty("user.home");
		} else {
			appDataDir = System.getProperty("user.home");
		}
		return new File(appDataDir, path);
	}
	
	public static void main(String[] args) {
		File file = new File("lib/natives");
		if (!file.exists()) file = new File("lib/natives-" + (System.getProperty("os.name").toLowerCase().contains("win") ? "windows" : "linux"));
		Configuration.LIBRARY_PATH.set(file.getAbsolutePath());
		
		new Main().run();
	}
	
}
