package hexagon.util;

import org.joml.Matrix4f;
import org.joml.Vector2f;
import org.joml.Vector3f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.model.HexagonModel;
import hexagon.world.Chunk;
import hexagon.world.World;

public class MathUtil {
	
	public static final Vector3f unitX = new Vector3f(1, 0, 0);
	public static final Vector3f unitY = new Vector3f(0, 1, 0);
	public static final Vector3f unitZ = new Vector3f(0, 0, 1);
	
	private MathUtil() {}
	
	public static Matrix4f transformationMatrix(Vector2f pos, Vector2f scale, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.identity();
		matr.translate(new Vector3f(pos.x, pos.y, 0));
		matr.scale(new Vector3f(scale.x, scale.y, 1));
		return matr;
	}
	
	public static Matrix4f transformationMatrix(Vector3f pos, Vector3f rotation, float scale, float globalScale, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.identity();
		Vector3f tempVec = new Vector3f(globalScale, globalScale, globalScale);
		if (globalScale != 1) matr.scale(tempVec);
		matr.translate(pos);
		matr.rotate((float) (Math.PI / 2d) - rotation.x, unitX);
		matr.rotate(rotation.y, unitY);
		matr.rotate(rotation.z, unitZ);
		tempVec.set(scale, scale, scale);
		matr.scale(tempVec);
		return matr;
	}
	
	public static Matrix4f projectionMatrix(float fov, float aspect, float near, float far, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.identity();
		matr.setPerspective(fov, aspect, near, far);
		return matr;
	}
	
	public static Matrix4f viewMatrix(Camera camera, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.identity();
		matr.rotate(camera.rotation.z, unitZ);
		matr.rotate(camera.rotation.x, unitX);
		matr.rotate(camera.rotation.y, unitY);
		return matr;
	}
	
	public static void chunkCoordsToReal(Vector3f chunkCoords, Vector3f dest) {
		float xx = chunkCoords.x * (HexagonModel.x60 + 1);
		dest.z = superMod(chunkCoords.z + chunkCoords.x * 0.5f, Chunk.ringSize) * 2f * HexagonModel.y60;
		dest.x = xx;
		dest.y = chunkCoords.y;
	}
	
	public static void realCoordsToChunk(Vector3f realCoords, Vector3f dest) {
		float xx = realCoords.x / (HexagonModel.x60 + 1);
		dest.z = superMod((realCoords.z / HexagonModel.y60 - xx) / 2f, Chunk.ringSize);
		dest.x = xx;
		dest.y = realCoords.y;
	}
	
	public static float hypot(float a, float b, float c) {
		return (float) Math.sqrt(a * a + b * b + c * c);
	}
	
	public static float clamp(float a, float low, float high) {
		return a < low ? low : a > high ? high : a;
	}
	
	public static int superMod(int x, int m) {
		x %= m;
		if (x < 0) x += m;
		return x;
	}
	
	public static float superMod(float x, float m) {
		x %= m;
		if (x < 0) x += m;
		if (x == m) x = 0;
		return x;
	}
	
	public static double superMod(double x, double m) {
		x %= m;
		if (x < 0) x += m;
		if (x == m) x = 0;
		return x;
	}
	
	public static int fitChunkZ(int z) {
		return superMod(z, Chunk.ringSize);
	}
	
	public static float fitChunkZ(float z) {
		return superMod(z, Chunk.ringSize);
	}
	
	public static Vector3f toBlockCoords(Vector3f vec, Vector3f result, Vector3f offset) {
		return toBlockCoords(vec, result, offset, true);
	}
	
	public static Vector3f toBlockCoords(Vector3f vec, Vector3f result, Vector3f offset, boolean adjustY) {
		if (result == null) result = new Vector3f();
		float camX = Main.main.camera.camX;
		double y = vec.y;
		double mult = adjustY ? Math.exp(y - Main.main.camera.camY) : 1;
		double x = ((vec.x - camX) / mult + camX) / (HexagonModel.x60 + 1);
		double z = vec.z / (2 * HexagonModel.y60) - x / 2;
		int xInt = (int) Math.round(x);
		int zInt = (int) Math.round(z);
		
		double xx, zz, xp, zp, wp;
		while (true) {
			xx = x - xInt;
			zz = z - zInt;
			if (Math.abs(xp = xx + 0.5 * zz) > 0.5) {
				if (xp > 0) xInt++;
				else xInt--;
			} else if (Math.abs(zp = zz + 0.5 * xx) > 0.5) {
				if (zp > 0) zInt++;
				else zInt--;
			} else if (Math.abs(wp = zp - xp) > 0.5) {
				if (wp > 0) {
					zInt++;
					xInt--;
				} else {
					zInt--;
					xInt++;
				}
			} else break;
		}
		
		xx = x - xInt;
		zz = z - zInt;
		xp = xx + 0.5 * zz;
		zp = zz + 0.5 * xx;
//		zp = yp - xp;//0.5 * (zz - xx);
		
		zInt = superMod(zInt, Chunk.totalSize);
/*		zInt %= Chunk.totalSize;
		if (zInt < 0) zInt += Chunk.totalSize;
		zInt %= Chunk.totalSize;*/
		
		if (offset != null) {
			result.set(xInt, (float) Math.floor(y), zInt);
			offset.set((float) xp, (float) y - result.y - 0.5f, (float) zp);
		} else result.set((float) x, (float) y, (float) z);
		return result;
	}
	
	public static Vec3f fromBlockCoords(Vector3f vec, Vector3f offset, Vec3f result) {
		if (result == null) result = new Vec3f();
		float camX = Main.main.camera.camX;
		double mult = Math.exp((vec.y + offset.y - Main.main.camera.camY) / World.radius);
		result.setCyl((float) (((HexagonModel.x60 + 1) * vec.x + offset.x - camX) * mult), World.radius * mult, (vec.x + 2 * vec.z + offset.z - Main.main.camera.position.z / HexagonModel.y60) * World.angleHalfHexagon);
		return result;
	}
	
	public static void floorAll(Vector3f vec) {
		vec.x = (float) Math.floor(vec.x);
		vec.y = (float) Math.floor(vec.y);
		vec.z = (float) Math.floor(vec.z);
	}
	
}
